import inspect
import sys
import traceback

from amqtt_db.base.base_plugin import BasePlugin
from amqtt_db.base.base_topic import BaseTopicEngine
from amqtt_db.base.constants import TOPIC_ENGINE, DB_CONNECT_STRING, DB_STRUCTURE
from amqtt_db.db.db_sa import SA
from amqtt_db.structure.structure import WideStructure


class DBPlugin(BasePlugin):
    """
    The DB Plugin
    """

    config_path = 'plugins.amqtt_db'
    db = None
    mapper = None
    topic_engine = None

    def __init__(self, context):
        super(DBPlugin, self).__init__(context)
        self.compose()

    def handle_exception(self, e):
        msg = "amqtt_db ERROR: {}".format(e.__repr__())
        self.context.logger.error(msg)
        exc_type, exc_value, exc_traceback = sys.exc_info()
        for line in traceback.format_tb(exc_traceback):
            self.context.logger.error(line[:-1])
        raise ImportError

    def compose(self):
        """
        Constructs the plugin by choosing the structure and the DB interface according to the config
        """
        self.get_topic_engine()
        self.get_db()
        self.get_mapper()
        self.register_events()

    def get_topic_engine(self):
        """
        Gets the topic engine according to the config
        """
        if TOPIC_ENGINE not in self.config:
            topic_engine = BaseTopicEngine
        else:
            topic_engine = self.config[TOPIC_ENGINE]

        try:
            self.topic_engine = topic_engine(self)
        except Exception as e:
            self.context.logger.error('No topic engine could be started')
            self.handle_exception(e)

    def get_db(self):
        """
        Gets the DB according to the config
        """
        if DB_CONNECT_STRING not in self.config:
            raise

        try:
            connect_string = self.config[DB_CONNECT_STRING]
        except KeyError as e:
            self.context.logger.error('No "db_connection" found in config file at: "{}"'.format(self.config_path))
            self.handle_exception(e)

        try:
            self.db = SA(self.context, connect_string)
        except Exception as e:
            msg = 'Connection to DB with connect string: "{}" failed'.format(connect_string)
            self.context.logger.error(msg)
            self.handle_exception(e)

    def get_mapper(self):
        """
        Gets the structure according to the config
        """
        try:
            _mapper_type = self.config[DB_STRUCTURE]  # noqa: F841
        except KeyError as e:
            msg = 'Cannot read/find "db_structure" entry in config file part "{}"'.format(self.config_path)
            self.context.logger.error(msg)
            self.handle_exception(e)

        try:
            self.mapper = WideStructure(self.topic_engine, self.db, self.context)  # ToDo: Mapper Factory
        except Exception as e:
            self.handle_exception(e)

    def register_events(self):
        events = {}
        method_names = inspect.getmembers(self.mapper, predicate=inspect.ismethod)
        for method_name, method in method_names:
            if method_name.startswith('on_'):
                events[method_name] = method
        return events

    def __getattr__(self, item):
        try:
            result = self.__getattribute__(item)
        except AttributeError:
            result = getattr(self.mapper, item)
        return result



import unittest
import aiounittest


from amqtt_db.plugin import DBPlugin
from amqtt_db.tests.resources.mockups import context, publish_packet


class TestPlugin(aiounittest.AsyncTestCase):

    def test_plugin(self):
        _plugin = DBPlugin(context)    # noqa: F841

    async def test_on_packet(self):
        plugin = DBPlugin(context)
        await plugin.on_mqtt_packet_received(packet=publish_packet)


if __name__ == '__main__':
    unittest.main()



import asyncio
import logging

from hbmqtt.broker import Broker
from hbmqtt.utils import read_yaml_config


def main():
    formatter = "[%(asctime)s] :: %(levelname)s - %(message)s"

    logging.basicConfig(level=logging.DEBUG, format=formatter)

    try:
        config = read_yaml_config('config.yaml')
    except FileNotFoundError:
        config = read_yaml_config('default_config.yaml')

    loop = asyncio.get_event_loop()
    broker = Broker(config)
    try:
        loop.run_until_complete(broker.start())
        loop.run_forever()
    except KeyboardInterrupt:
        loop.run_until_complete(broker.shutdown())
    finally:
        loop.close()


if __name__ == "__main__":
    main()



from amqtt_db.base.base_topic import BaseTopicEngine
from amqtt_db.base.errors import TopicNotFound


class TopicEngine(BaseTopicEngine):

    def topic2handler(self, topic):
        """
        Get a handler for a topic according to the payload definitions
        :param topic:
        :return: The handler
        """
        if topic in self.topic_handlers:
            return self.topic_handlers[topic]
        else:
            result = self.match_topic(topic)
            if result is not None:
                self.topic_handlers[topic] = result
                return result
        raise TopicNotFound()

    def match_topic(self, topic):
        """
        Match a topic to an handler according to the config
        :param topic:
        :return: a topic handler
        """




from unittest.mock import MagicMock

import yaml
from hbmqtt.broker import BrokerContext
from hbmqtt.mqtt import PublishPacket
from hbmqtt.mqtt.publish import PublishPayload, PublishVariableHeader

import logging

CONFIG = """
plugins:
  - amqtt_db

amqtt_db:
  db_structure : amqtt_db.structure.structure:WideStructure
  db_connection : 'sqlite:///test_topic.qlite'

  payload :
    hall/temp_rh :
      amqtt_db.payload.json_decoder.JSONDecoder :
        amqtt_db.payload.flat_deserializer.FlatDeserializer :
          temp : float
          rh : float
          press : float
          gas : float
          'current time' : datetime.datetime.utcfromtimestamp
"""


context = MagicMock(spec=BrokerContext)()
context.config = yaml.load(CONFIG, Loader=yaml.FullLoader)
context.logger = logging.getLogger('test')


publish_packet = PublishPacket()
publish_packet.variable_header = PublishVariableHeader('hall/temp_rh')

payload = PublishPayload()
payload.data = bytearray(b'''{"3c610515b7c4": {"current time": 1617663897, "gas": 46.922, "temp": 24.35012, "press": 936.7473, "rh": 29.88471}}''')

publish_packet.payload = payload



from amqtt_db.base.constants import ON_SAVE_SESSION, ON_FIND_SESSION, ON_DEL_SESSION, ON_BROKER_POST_SHUTDOWN, \
    ON_MQTT_PACKET_RECEIVED, BASE_CONFIG_PATH
from amqtt_db.base.delegator import Delegator


class BasePlugin(Delegator):
    """
    Plugin base class
    """
    DELEGATED_METHODS = {
        'structure': [
            ON_SAVE_SESSION,
            ON_FIND_SESSION,
            ON_DEL_SESSION,
            ON_BROKER_POST_SHUTDOWN,
            ON_MQTT_PACKET_RECEIVED,
            ]
    }

    config_path = BASE_CONFIG_PATH   # dotted name path e.g 'plugins.timescaledb'
    config = None

    def __init__(self, context):
        self.context = context
        self.init_config(context)

    def init_config(self, context):
        """
        Use the config_path to derive the config portion of this test_plugin and store it under self.config.
        :param context:
        :return:
        """
        try:
            temp_config = context.config
            path_parts = self.config_path.split('.')
            for path_part in path_parts[:-1]:
                temp_config = temp_config[path_part]
            if path_parts[-1] in temp_config:
                self.config = context.config[path_parts[-1]]
            else:
                raise ImportError()

        except KeyError as e:
            msg = 'amqtt_db: Plugin config not found under "{}" in the config file '.format(self.config_path)
            self.context.logger.error(msg)
            raise ImportError(e)

    def register_events(self):
        """
        This method is called from the test_plugin manager to register the events
        """



from datetime import date, datetime, time

from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.orm import sessionmaker

from migrate.versioning.schema import Table as MiTable, Column as MiColumn

from amqtt_db.base.base_db import BaseDB

import sqlalchemy as sa

from amqtt_db.base.errors import NoSAMappingForType
from amqtt_db.base.type_mapper import BaseTypeMapper
from amqtt_db.db.base import Base


class SATypeMapper(BaseTypeMapper):
    """Maps python types to SQLAlchemy Column types"""

    def __init__(self):
        super(SATypeMapper, self).__init__()
        _map = {
            float: sa.FLOAT,
            int: sa.BigInteger,
            str: sa.VARCHAR,
            date: sa.types.Date,
            time: sa.types.Time,
            datetime: sa.types.DateTime,
        }
        self.map.update(_map)


class SA(BaseDB):
    """
    SQLAlchemy storage
    """
    autocommit = None
    connect_string = None
    engine = None
    session = None
    type_mapper = None

    def init_db(self, connect_string, autocommit=False, echo=False):
        """
        Initialize the DB
        """
        self.autocommit = autocommit
        self.connect_string = connect_string
        self.engine = create_engine(connect_string, echo=echo)
        self.session = sessionmaker(bind=self.engine, autocommit=self.autocommit)()
        self.type_mapper = SATypeMapper()

    def create_table(self, name, column_def):
        """
        Dynamically create a table from a column definition
        :param name: Table name
        :param column_def: Column definition, dict like
        """
        # The blueprint of the table
        table_description = {
            '__tablename__': name,
            'id': sa.Column(sa.BigInteger().with_variant(sa.Integer, "sqlite"), primary_key=True),
            'sqlite_autoincrement': True,
        }
        # The columns from the column_def are converted to SA Types and added to the blueprint
        for col_name, col_type in column_def.items():
            try:
                sa_col_type = self.type_mapper.map[col_type]
            except KeyError:
                raise NoSAMappingForType('Python type "{}" has no SA mapping defined'.format(col_type))
            table_description[col_name] = sa.Column(sa_col_type)

        # Construct the Table from the blueprint
        _table = type(name, (Base,), table_description)  # noqa: F841
        # Create the table in the DB
        Base.metadata.create_all(self.engine, tables=[Base.metadata.tables[name]])
        # Add the table to the metadata
        Base.metadata._add_table(name, None, Base.metadata.tables[name])

    def handle_new_columns(self, topic_cls, data):
        """
        Handle new columns.
        :param topic_cls: The topic decl class
        :param data: The data from the payload
        :return:
        """
        # Check if we really need additional columns
        new_cols = self.find_new_colums(topic_cls, data)
        # No new columns we return
        if len(new_cols) == 0:
            return
        # We need new columns, add them
        self.add_new_columns(topic_cls, new_cols)

    def find_new_colums(self, topic_cls, data):
        """
        This is good place to do some heuristics to check if really new columns should be made.
        :param topic_cls: The topic decl class
        :param data: The data from the payload
        :return: List of new columns if any
        """

        new_columns = {}
        existing_columns = topic_cls.__mapper__.columns
        for col_name, value in data.items():
            if col_name not in existing_columns:
                col_type = type(value)
                new_columns[col_name] = self.type_mapper.map[col_type]
        return new_columns

    def add_new_columns(self, topic_cls, column_def):
        """
        This is a bit of a hack since SQLAlchemy does not come with an out of the box solution
        for such a volatile DB usage we need. SQLAlchemy is mostly used for quite _static DB schema where migrations
        happen infrequently. At the IoT-Front we have to deal with permanent changes of the DB schema due to changes
        in the messages.
        """
        # The first part utilizes SQLalchemy.migrate to add new columns to the topic table.
        table_name = str(topic_cls.__table__.name)
        # Create a migration table.
        table = MiTable(table_name, MetaData(bind=self.engine))
        # for any new column
        for col_name, col_type in column_def.items():
            # Create a Migration Column
            col = MiColumn(col_name, col_type)
            # Do the migration
            col.create(table)

        # Now that we have a new table in the DB we should get rid of the old topic table,
        # so we remove it from the metadata.
        Base.metadata.remove(topic_cls.__table__)

        # The new Topic Table we get by inspecting the the Table from the DB correctly
        new_table = Table("test/topic_growth", Base.metadata, autoload_with=self.engine)

        # From the new Topic Table we generate its declarative class
        new_dcl = type(str(table_name), (Base,), {'__table__': new_table})

        # Since the new topic table was gained by inspection and not declarative construction our nice mechanism
        # (see db.base.py) to link the declarative class to the table automatically, failed.
        # so we set the link manually
        new_table.decl_class = new_dcl

        # At last we add the new topic table to the metadata.
        Base.metadata._add_table(table_name, None, new_table)

    def get_column_def(self, data):
        """
        Get a cloumn definition from data.
        :param data: Data from the payload
        :return: Column definition with the types of the data
        """
        result = {}
        for col_name, col_data in data.items():
            result[col_name] = type(col_data)
        return result

    def create_topic_table(self, topic, data):
        """
        Create a new topic table. At frist get the column definition. Then create the table from it.
        :param topic: MQTT Topic
        :param data: Data from the Payload
        """
        self.logger.debug('Building new table for topic {}'.format(topic))
        column_def = self.get_column_def(data)
        self.create_table(topic, column_def)

    def get_topic_cls(self, topic, data):
        """
        If the topic cannot be found in the metadata a new topic table will be creatd.
        For this process we need the current data to determine the columns needed.
        columns to be created.
        :param topic: The mqtt topic
        :param data: The data from the payload.
        :return: the declarative class for the topic
        """

        if topic not in Base.metadata.tables:
            self.create_topic_table(topic, data)

        topic_table = Base.metadata.tables[topic]

        return topic_table.decl_class

    def add_packet(self, session, sender, topic, data):
        """
        Insert a package into to DB.
        At first find the topic declarative class. Then instantiate it with the data from the payload.
        Add it to the session and commit.
        If a TypeError occurs we add the missing columns and try again.
        :param session: Not used
        :param sender: The sender ID of the package
        :param topic: The MQTT topic
        :param data: The data from the payload
        """

        topic_cls = self.get_topic_cls(topic, data)
        try:
            topic_line = topic_cls(**data)
        except TypeError:
            self.handle_new_columns(topic_cls, data)
            topic_cls = self.get_topic_cls(topic, data)
            topic_line = topic_cls(**data)

        self.session.add(topic_line)
        self.session.commit()

    def query_topic(self, topic, data):
        topic_cls = self.get_topic_cls(topic, data)

        query = self.session.query(topic_cls)
        for col_name, value in data.items():
            query = query.filter(getattr(topic_cls, col_name) == value)

        return query




from amqtt_db.base.base_deserializer import BaseDeserializer
from amqtt_db.base.utils import get_class_func_by_name


class FlatDeserializer(BaseDeserializer):

    def __init__(self, types_strs):
        """
        Flat deserialize is given a column_name -> type_str mapping.
        It derives a mapping column_name -> type (Class/Function(Factory))
        :param types_strs: column_name -> types_str mapping

        >>> fs = FlatDeserializer({'a_datetime': 'datetime.datetime.fromtimestamp'})
        >>> type(fs.col2type['a_datetime'])
        <class 'builtin_function_or_method'>

        >>> fs = FlatDeserializer({'a_float': 'float'})
        >>> fs.col2type['a_float']
        <class 'float'>

        >>> fs = FlatDeserializer({'a_float': 'floata'})
        >>> fs.col2type['a_float']
        <class 'float'>

        """

        self.type_strs = types_strs
        self.col2type = {}    # Hold the actual mapping
        for key, type_str in types_strs.items():
            self.col2type[key] = (get_class_func_by_name(type_str))  # get the string to class/function mapping

    def deserialize(self, data):
        """
        Deserialize by key name assuming a flat data structure.
        :param data: data
        :return: typed data
        """
        result = {}
        for key, value in data.items():
            result[key] = self.col2type[key](value)  # do the mapping
        return result



"""
Base deserializer class. The deserializer unpacks the payload and maps it to DB compatible types.
"""


class BaseDeserializer(object):

    def deserialize(self, data):
        """
        Map the data in data to types. This is done according to the given structure for this particular
        data.
        :param data: The incoming data
        :return: a structure of the form {value:type}
        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['amqtt_db',
 'amqtt_db.base',
 'amqtt_db.db',
 'amqtt_db.payload',
 'amqtt_db.structure',
 'amqtt_db.tests',
 'amqtt_db.tests.resources',
 'amqtt_db.tests.test_db',
 'amqtt_db.tests.test_plugin']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy>=1.4.1,<2.0.0',
 'Sphinx>=3.5.3,<4.0.0',
 'aiounittest>=1.4.0,<2.0.0',
 'amqtt>=0.10.0-alpha.3,<0.11.0',
 'coverage>=5.5,<6.0',
 'psycopg2>=2.8.6,<3.0.0',
 'sphinx-rtd-theme>=0.5.2,<0.6.0',
 'sqlalchemy-migrate>=0.13.0,<0.14.0']

entry_points = \
{'amqtt.broker.plugins': ['amqtt_db = amqtt_db.plugin:DBPlugin']}

setup_kwargs = {
    'name': 'amqtt-db',
    'version': '0.1.2',
    'description': '',
    'long_description': 'amqtt_db\n========\n\n![license](https://img.shields.io/github/license/volkerjaenisch/amqtt_db?style=flat-square)\n![travis](https://api.travis-ci.org/volkerjaenisch/amqtt_db.svg?branch=main)\n![coverals](https://coveralls.io/repos/github/volkerjaenisch/amqtt_db/badge.svg)\n[![PyPI](https://img.shields.io/pypi/v/amqtt_db?style=flat-square)](https://pypi.org/project/amqtt/)\n[![Documantation](https://img.shields.io/readthedocs/amqtt-db.svg)](https://amqtt_db.readthedocs.io/en/latest/)\n\nDB persistence for amqtt.\n\nObjective\n---------\n\namqtt_db will persist payloads received by the [amqtt broker](https://github.com/Yakifo/amqtt) into performant relational databases.\nSQLAlchemy as well as timescaleBD are the target RMDB-Systems.\n\namqtt_db will do four steps to persist the amqtt data:\n\n 1) decoding the payload (e.G. from binary, JSON or which ever encoding)\n 1) deserializing the payload to typed Python entities\n 1) structure the session, topic, property, value information into a relational model of your choice\n 1) generate the necessary tables, columns to store the data \n\nAll of this steps can be configured via the amqtt yaml config. And you can even replace any these steps for each topic \nby your code in terms of Python plugins.\namqtt is designed to be enhanced and extended.\n\n\nPerformance\n-----------\n\nFlexibility comes with a penalty on performance. The more layers of classes and filters we \nimplement the higher the performance penalty.   \n\nSo we optimize the data flow by an optimistic approach. \n\namqtt_db expects that the decoding, deserializing, transformations, target DB, target tables, table colums \netc. are all well in place if it deals with a single incoming packet.\nIf the handling of that package fails, exceptions will be raised, and the error handling rushes in to deal with the problem.\n\nSince the change rate on the decoding, deserializing, database model is quite low this optimistic approach will be quite performant. \n\nDocumentation\n-------------\n\nPlease have a look at the [documentation](http://amqtt-db.readthedocs.io).\n',
    'author': 'volker',
    'author_email': 'volker.jaenisch@inqbus.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/volkerjaenisch/amqtt_db',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)





class TopicNotFound(Exception):
    """
    Raised if a topic is not found
    """


class TypeNotFound(Exception):
    """
    Raised if a type is not found
    """


class NoSAMappingForType(Exception):
    """
    Raised if no SA mapping for a given python type is found
    """
